package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import okhttp3.Headers;
import okhttp3.Interceptor;
import okhttp3.MediaType;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;

import java.io.IOException;
import java.net.Proxy;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;

public class HttpClient {

    private final OkHttpClient httpClient;
    private final ObjectMapper objectMapper = new ObjectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    private final boolean throwExceptionsOnFailure;

    public HttpClient(boolean throwExceptionsOnFailure) {
        this.httpClient = new OkHttpClient();
        this.throwExceptionsOnFailure = throwExceptionsOnFailure;
    }

    public HttpClient(boolean throwExceptionsOnFailure, List<Interceptor> interceptors) {
        this.httpClient = builderWithInterceptors(interceptors).build();
        this.throwExceptionsOnFailure = throwExceptionsOnFailure;
    }

    public HttpClient(boolean throwExceptionsOnFailure, List<Interceptor> interceptors, Proxy proxy) {
        this.httpClient = builderWithInterceptors(interceptors).proxy(proxy).build();
        this.throwExceptionsOnFailure = throwExceptionsOnFailure;
    }

    private OkHttpClient.Builder builderWithInterceptors(List<Interceptor> interceptors) {

        OkHttpClient.Builder builder = new OkHttpClient.Builder();
        interceptors.forEach(builder::addInterceptor);

        return builder;
    }

    public <R> HttpResponse<R> get(String url, Class<R> responseType) throws IOException {
        return get(url, Map.of(), responseType);
    }

    public <R> HttpResponse<R> get(String url, Map<String, String> headers, Class<R> responseType) throws IOException {

        Request request = new Request.Builder()
                .url(url)
                .get()
                .headers(Headers.of(headers))
                .build();

        try (Response response = this.httpClient.newCall(request).execute()) {
            return this.evaluateStatusAndRespond(response, responseType);
        }
    }

    public <P, R> HttpResponse<R> post(String url, P httpPayload, Class<R> responseType) throws IOException {
        return post(url, Map.of(), httpPayload, responseType);
    }

    public <P, R> HttpResponse<R> post(String url, Map<String, String> headers, P httpPayload, Class<R> responseType) throws IOException {

        RequestBody requestBody = this.createRequestBody(headers, httpPayload);
        Request request = new Request.Builder()
                .url(url)
                .post(requestBody)
                .headers(Headers.of(headers))
                .build();

        try (Response response = this.httpClient.newCall(request).execute()) {
            return this.evaluateStatusAndRespond(response, responseType);
        }
    }

    public <P, R> HttpResponse<R> put(String url, P httpPayload, Class<R> responseType) throws IOException {
        return put(url, Map.of(), httpPayload, responseType);
    }

    public <P, R> HttpResponse<R> put(String url, Map<String, String> headers, P httpPayload, Class<R> responseType) throws IOException {

        RequestBody requestBody = this.createRequestBody(headers, httpPayload);
        Request request = new Request.Builder()
                .url(url)
                .put(requestBody)
                .headers(Headers.of(headers))
                .build();

        try (Response response = this.httpClient.newCall(request).execute()) {
            return this.evaluateStatusAndRespond(response, responseType);
        }
    }

    private <T> RequestBody createRequestBody(Map<String, String> headers, T httpPayload) throws JsonProcessingException {

        byte[] body;
        if (httpPayload.getClass().equals(String.class)) {
            body = ((String) httpPayload).getBytes(StandardCharsets.UTF_8);
        } else {
            body = this.objectMapper.writeValueAsString(httpPayload).getBytes(StandardCharsets.UTF_8);
        }

        if (headers.containsKey(HttpHeaders.CONTENT_TYPE)) {
            return RequestBody.create(body, MediaType.parse(headers.get(HttpHeaders.CONTENT_TYPE)));
        }

        return RequestBody.create(body, MediaType.get(MimeTypes.APPLICATION_JSON));
    }

    private <R> HttpResponse<R> evaluateStatusAndRespond(Response response, Class<R> responseType) throws IOException {

        if (this.throwExceptionsOnFailure && !response.isSuccessful()) {
            throw new RestApiException("HTTP request failed.", response.code());
        }

        ResponseBody responseBody = response.body();
        if (responseBody == null) {
            throw new IOException("Response body is null.");
        }

        if (responseType.equals(Void.class)) {
            return new HttpResponse<>(response.code(), null);
        }

        if (responseType.equals(String.class)) {
            return new HttpResponse<>(response.code(), (R) responseBody.string());
        }

        return new HttpResponse<>(response.code(), this.objectMapper.readValue(responseBody.string(), responseType));
    }
}
